// Copyright Cromwell D. Enage 2017.
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

#ifndef BOOST_PARAMETER_AUX_PREPROCESSOR_SEQ_MERGE_HPP
#define BOOST_PARAMETER_AUX_PREPROCESSOR_SEQ_MERGE_HPP

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_0(seq0, seq1) seq1
/**/

#include <boost/preprocessor/seq/seq.hpp>
#include <boost/preprocessor/seq/push_front.hpp>

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_1(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT(seq1, BOOST_PP_SEQ_HEAD(seq0))
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_2(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_1(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_3(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_2(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_4(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_3(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_5(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_4(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_6(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_5(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_7(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_6(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_8(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_7(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_9(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_8(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_10(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_9(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_11(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_10(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_12(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_11(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_13(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_12(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_14(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_13(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_15(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_14(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_16(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_15(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_17(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_16(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_18(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_17(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_19(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_18(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_20(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_19(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_21(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_20(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_22(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_21(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_23(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_22(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_24(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_23(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_25(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_24(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_26(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_25(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_27(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_26(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_28(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_27(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_29(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_28(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_30(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_29(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_31(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_30(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_32(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_31(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_33(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_32(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_34(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_33(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_35(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_34(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_36(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_35(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_37(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_36(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_38(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_37(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_39(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_38(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_40(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_39(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_41(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_40(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_42(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_41(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_43(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_42(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_44(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_43(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_45(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_44(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_46(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_45(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_47(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_46(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_48(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_47(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_49(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_48(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_50(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_49(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_51(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_50(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_52(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_51(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_53(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_52(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_54(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_53(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_55(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_54(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_56(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_55(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_57(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_56(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_58(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_57(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_59(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_58(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_60(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_59(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_61(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_60(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_62(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_61(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_63(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_62(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_64(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_63(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_65(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_64(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_66(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_65(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_67(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_66(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_68(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_67(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_69(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_68(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_70(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_69(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_71(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_70(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_72(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_71(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_73(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_72(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_74(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_73(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_75(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_74(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_76(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_75(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_77(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_76(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_78(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_77(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_79(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_78(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_80(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_79(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_81(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_80(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_82(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_81(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_83(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_82(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_84(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_83(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_85(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_84(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_86(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_85(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_87(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_86(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_88(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_87(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_89(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_88(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_90(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_89(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_91(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_90(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_92(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_91(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_93(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_92(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_94(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_93(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_95(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_94(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_96(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_95(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_97(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_96(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_98(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_97(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_99(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_98(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_100(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_99(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_101(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_100(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_102(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_101(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_103(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_102(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_104(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_103(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_105(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_104(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_106(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_105(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_107(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_106(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_108(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_107(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_109(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_108(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_110(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_109(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_111(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_110(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_112(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_111(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_113(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_112(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_114(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_113(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_115(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_114(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_116(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_115(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_117(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_116(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_118(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_117(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_119(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_118(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_120(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_119(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_121(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_120(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_122(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_121(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_123(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_122(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_124(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_123(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_125(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_124(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_126(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_125(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_127(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_126(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_128(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_127(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_129(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_128(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_130(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_129(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_131(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_130(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_132(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_131(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_133(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_132(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_134(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_133(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_135(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_134(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_136(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_135(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_137(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_136(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_138(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_137(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_139(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_138(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_140(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_139(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_141(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_140(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_142(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_141(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_143(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_142(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_144(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_143(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_145(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_144(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_146(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_145(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_147(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_146(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_148(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_147(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_149(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_148(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_150(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_149(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_151(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_150(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_152(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_151(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_153(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_152(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_154(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_153(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_155(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_154(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_156(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_155(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_157(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_156(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_158(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_157(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_159(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_158(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_160(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_159(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_161(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_160(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_162(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_161(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_163(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_162(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_164(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_163(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_165(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_164(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_166(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_165(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_167(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_166(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_168(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_167(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_169(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_168(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_170(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_169(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_171(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_170(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_172(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_171(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_173(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_172(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_174(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_173(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_175(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_174(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_176(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_175(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_177(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_176(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_178(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_177(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_179(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_178(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_180(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_179(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_181(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_180(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_182(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_181(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_183(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_182(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_184(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_183(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_185(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_184(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_186(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_185(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_187(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_186(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_188(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_187(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_189(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_188(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_190(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_189(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_191(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_190(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_192(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_191(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_193(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_192(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_194(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_193(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_195(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_194(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_196(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_195(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_197(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_196(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_198(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_197(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_199(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_198(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_200(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_199(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_201(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_200(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_202(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_201(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_203(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_202(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_204(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_203(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_205(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_204(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_206(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_205(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_207(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_206(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_208(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_207(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_209(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_208(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_210(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_209(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_211(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_210(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_212(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_211(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_213(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_212(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_214(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_213(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_215(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_214(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_216(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_215(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_217(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_216(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_218(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_217(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_219(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_218(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_220(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_219(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_221(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_220(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_222(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_221(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_223(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_222(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_224(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_223(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_225(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_224(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_226(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_225(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_227(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_226(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_228(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_227(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_229(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_228(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_230(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_229(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_231(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_230(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_232(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_231(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_233(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_232(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_234(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_233(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_235(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_234(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_236(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_235(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_237(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_236(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_238(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_237(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_239(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_238(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_240(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_239(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_241(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_240(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_242(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_241(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_243(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_242(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_244(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_243(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_245(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_244(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_246(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_245(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_247(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_246(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_248(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_247(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_249(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_248(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_250(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_249(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_251(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_250(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_252(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_251(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_253(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_252(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_254(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_253(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE_255(seq0, seq1) \
    BOOST_PP_SEQ_PUSH_FRONT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_254(BOOST_PP_SEQ_TAIL(seq0), seq1) \
      , BOOST_PP_SEQ_HEAD(seq0) \
    )
/**/

#include <boost/preprocessor/seq/size.hpp>
#include <boost/preprocessor/cat.hpp>

#define BOOST_PARAMETER_AUX_PP_SEQ_MERGE(seq0, seq1) \
    BOOST_PP_CAT( \
        BOOST_PARAMETER_AUX_PP_SEQ_MERGE_, BOOST_PP_SEQ_SIZE(seq0) \
    )(seq0, seq1)
/**/

#endif  // include guard

